from src.models.request import Request
from src.preprocessing import evaluate

from rich import inspect


class Step:
    def __init__(
        self, 
        step: dict,
        steps: dict,
    ):
        for key in step:
            step[key] = evaluate(step[key], steps)
        
        self.name = step['name']
        self.operation_id = step['operation_id']
        self.method = step['method']
        self.path = step['path']
        self.status_code = step['status_code']
        
        if 'headers' in step:
            self.headers = step['headers'] 
        else:
            self.headers = None
        
        if 'body' in step:
            self.body = step['body'] 
        else:
            self.body = None

        if 'raw_body' in step:
            self.body = step['raw_body']
        
        if 'params' in step:
            self.params = step['params'] 
        else:
            self.params = None

        if 'auth' in step:
            self.auth = step['auth'] 
        else:
            self.auth = None

    def construct_request(self, base_url, global_auth):
        return Request(
                base_url=base_url,
                method=self.method,
                path=self.path,
                params=self.params,
                headers=self.headers,
                global_auth=global_auth,
                auth=self.auth,
                body=self.body
            )



import pytest
from hamcrest import assert_that, is_, has_items, instance_of

from src.parsing import *

class TestParsing:
    """Class for basic unit testing of the parsing.py module
    """
    steps_file_path="./tests/fixtures/valid/steps/person_api.yml"
    spec_file_path = "./tests/fixtures/valid/specs/person_api.yml"

    invalid_steps_missing_operation_id_file_path="./tests/fixtures/invalid/steps/missing_property.yml"
    invalid_spec_missing_openapi_property_file_path = "./tests/fixtures/invalid/specs/invalid_spec_1.yml"

    test_steps=parse_steps_file(steps_file_path)

    def test_parse_steps_file_with_valid_steps(self):
        """Ensure a step file is parsed correctly
        """
        steps = parse_steps_file(self.steps_file_path)

        assert_that(steps, is_(instance_of(dict)))
        assert_that(steps, has_items('coverage_threshold', 'base_url', 'steps'))
        assert_that(steps['coverage_threshold'], is_(instance_of(float)))
        assert_that(steps['base_url'], is_(instance_of(str)))
        assert_that(steps['steps'], is_(instance_of(list)))

        assert_that(steps['steps'][0], has_items('name', 'operation_id', 'method', 'path', 'headers', 'body', 'auth', 'status_code'))
        assert_that(steps['steps'][0]['name'], is_(instance_of(str)))
        assert_that(steps['steps'][0]['operation_id'], is_(instance_of(str)))
        assert_that(steps['steps'][0]['method'], is_(instance_of(str)))
        assert_that(steps['steps'][0]['path'], is_(instance_of(str)))
        assert_that(steps['steps'][0]['headers'], is_(instance_of(dict)))
        assert_that(steps['steps'][0]['body'], is_(instance_of(dict)))
        assert_that(steps['steps'][0]['auth'], is_(instance_of(dict)))
        assert_that(steps['steps'][0]['status_code'], is_(instance_of(int)))

        assert_that(steps['steps'][1], has_items('name', 'operation_id', 'method', 'path', 'headers', 'auth', 'status_code'))

    def test_parse_steps_file_with_missing_operation_id(self):
        """Ensure that a missing required property raises a ValidationError
            ValidationError comes from the jsonschema library
        """
        with pytest.raises(ValidationError) as e:
            invalid_steps = parse_steps_file(self.invalid_steps_missing_operation_id_file_path)

    def test_parse_valid_spec_file(self):
        """Ensure an OpenAPI document is parsed correctly
        """
        spec, operation_schemas = parse_spec_file(self.test_steps, self.spec_file_path)
        assert_that(spec, is_(instance_of(dict)))
        # TODO: More tests to assert children are properly formatted
        assert_that(operation_schemas, is_(instance_of(dict)))
        # TODO: More tests

    def test_parse_invalid_spec_file(self):
        """Ensure an OpenAPI document is parsed correctly
            OpenAPIValidationError comes from openapi_spec_validator library
        """
        with pytest.raises(OpenAPIValidationError):
            spec, operation_schemas = parse_spec_file(self.test_steps, self.invalid_spec_missing_openapi_property_file_path)



import os
import json

import yaml
from jsonschema import validate, ValidationError
from json_ref_dict import materialize, RefDict
from openapi_spec_validator import validate_spec
from openapi_spec_validator.exceptions import OpenAPIValidationError
from rich import print

from .errors import *

def parse_steps_file(step_file_path: str) -> dict:
    """Parse a .yml steps file and convert it to a python dictionary

    Args:
        step_file_path (str): Relative path to steps file

    Raises:
        FileNotFoundError: Step file not found
        ValidationError: Step file not valid

    Returns:
        dict: Step file converted to a python dictionary
    """
    try:
        steps_schema_path = os.path.join(os.path.dirname(__file__), "steps_schema.yml")
        with open(steps_schema_path, "r") as step_schema_file:
            steps_schema = yaml.safe_load(step_schema_file.read())
    except FileNotFoundError as e:
        raise FileNotFoundError("Steps schema file not found") from e

    try:
        with open(step_file_path, "r") as step_file:
            steps = yaml.safe_load(step_file.read())
            validate(instance=steps, schema=steps_schema)

    except ValidationError as e:
        raise ValidationError(f"Steps file has the following syntax errors: {e.message} ") from e
    except FileNotFoundError as e:
        raise FileNotFoundError(f"Steps file {step_file_path} not found") from e

    print("[bold green]--Successfully Validated Steps File--[/bold green]")
    return steps


def parse_spec_file(steps: dict, spec_file_path: str) -> tuple:
    """Parse a valid OpenAPI document into a python dictionary

    Args:
        spec_file_path (str): Relative path to OpenAPI spec file

    Raises:
        OpenAPIValidationError: Invalid API Definition
        FileNotFoundError: Spec file not found
        UnsupportedSchemaError: Schema not supported

    Returns:
        tuple: Returns a tuple containing two dictionaries - the parsed API spec and
            a dictionary with all of the openapi operations for quick reference
    """
    try:
        spec = materialize(RefDict(spec_file_path))
        validate_spec(spec)
    except OpenAPIValidationError as e:
        raise OpenAPIValidationError(f"API Spec file has the following syntax errors: {e.message} ") from e
    except FileNotFoundError as e:
        raise FileNotFoundError(f"API Spec file {spec_file_path} not found") from e

    # Get list of steps operationIds
    steps_opertaion_id_list = []
    for step in steps['steps']:
        steps_opertaion_id_list.append(step['operation_id'])

    # Create operation schemas dict for easier parsing
    operation_schemas: dict = {}
    for path in spec['paths']:
        for method in spec['paths'][path]:
            op_id = spec['paths'][path][method]['operationId']
            if op_id in steps_opertaion_id_list:
                operation = spec['paths'][path][method]
                operation_schemas[op_id] = {}

                # Parse and Validate Request Bodies
                if 'requestBody' in operation.keys():
                    # TODO: Support multiple content types if one of them is JSON
                    if len(operation['requestBody']['content']) > 1:
                        if 'application/json' not in operation['responses'][status_code]['content']:
                            raise UnsupportedSchemaError(f"There are too many request body content types for"
                                                                f" the operation: {op_id}")
                    if 'application/json' in operation['requestBody']['content'].keys():
                        operation_schemas[op_id]['requestBody'] = \
                            operation['requestBody']['content']['application/json']['schema']
                    elif 'application/octet-stream' in operation['requestBody']['content'].keys():
                        operation_schemas[op_id]['requestBody'] = \
                            operation['requestBody']['content']['application/octet-stream']['schema']
                    else:
                        raise UnsupportedSchemaError(f"request body content type: "
                                                    f"{list(operation['requestBody']['content'].keys())[0]}"
                                                    f" unsupported for the operation: {op_id}")

                # Parse and Validate Response Bodies
                operation_schemas[op_id]['responses'] = {}
                for status_code in operation['responses'].keys():
                    if 'content' in operation['responses'][status_code].keys():
                        # TODO: Support multiple content types if one of them is JSON
                        if len(operation['responses'][status_code]['content']) > 1:
                            if 'application/json' not in operation['responses'][status_code]['content']:
                                raise UnsupportedSchemaError(f"There are too many response body content types for"
                                                            f" the operation: {op_id}")
                        if 'application/json' in operation['responses'][status_code]['content'].keys():
                            operation_schemas[op_id]['responses'][status_code] = \
                                operation['responses'][status_code]['content']['application/json']['schema']
                        elif 'application/octet-stream' in operation['responses'][status_code]['content'].keys():
                            operation_schemas[op_id]['responses'][status_code] = \
                                operation['responses'][status_code]['content']['application/octet-stream']['schema']
                        else:
                            raise UnsupportedSchemaError(f"response body content type: "
                                                         f"{list(operation['responses'][status_code]['content'].keys())[0]}"
                                                         f" unsupported for the operation: {op_id}")
                    else:
                        operation_schemas[op_id]['responses'][status_code] = operation['responses'][status_code]

    print("[bold green]--Successfully Validated Spec File--[/bold green]")
    return spec, operation_schemas



import requests
from requests.auth import HTTPBasicAuth
from rich import inspect
from typing import Union


from src.models.response import Response

class Request:
    def __init__(
        self,
        base_url: str,
        method: str,
        path: str,
        params: dict,
        headers: dict,
        global_auth: dict,
        auth: dict,
        body: Union[dict, str],
    ):
        self.base_url = base_url
        self.method = method
        self.path=path
        self.params=params
        self.headers=headers
        self.global_auth=global_auth
        self.auth=auth
        self.body=body

        if not self.auth:
            if not self.global_auth:
                self.auth = {'username': '', 'password': ''}
            else:
                self.auth=self.global_auth

    def send(self):
        if type(self.body) == str:
            r = requests.request(
                url=self.base_url + self.path,
                method=self.method,
                params=self.params,
                headers=self.headers,
                # TODO: Beef up this logic to ensure a strict data type (binary/bytes?)
                data=self.body,
                auth=HTTPBasicAuth(self.auth['username'], self.auth['password']),
            )
            return Response(
                status_code=r.status_code,
                headers=r.headers,
                data=str(r.text)
            )
        else:
            r = requests.request(
                    url=self.base_url + self.path,
                    method=self.method,
                    params=self.params,
                    headers=self.headers,
                    json=self.body,
                    auth=HTTPBasicAuth(self.auth['username'], self.auth['password']),
            )
            return Response(
                status_code=r.status_code,
                headers=r.headers,
                data=str(r.text)
            )




from .preprocessing import evaluate
from .models import Step
from .verify import *

from rich import print, print_json
import json

def make_requests(steps_data: dict, operation_schemas: dict, fail_fast: bool, verbose: bool):

    base_url: str = steps_data["base_url"]

    # Set global auth if it exists in the step file
    global_auth: dict = None
    if 'auth' in steps_data.keys():
        global_auth: dict = {}
        for key, value in steps_data["auth"].items():
            global_auth[key] = evaluate(value)
    else:
        global_auth = None

    # Get steps list
    steps_data: list = steps_data["steps"]

    # Create Global Steps List
    steps: dict = {}

    for s in steps_data:
        print(f"Step Name: {s['name']}")
        step = Step(s, steps)
        request = step.construct_request(base_url, global_auth)

        try:
            response_schema = operation_schemas[s['operation_id']]['responses'][str(s['status_code'])]
        except KeyError as e:
            print("[bold red]Response Validation Error[/bold red]")
            raise KeyError(f"Response code of {e} not found in responses for {s['operation_id']}") from e

        if 'requestBody' in operation_schemas[s['operation_id']].keys():
            verify_request_body(request.body, operation_schemas[s['operation_id']]['requestBody'])

        response = request.send()
        if verbose:
            if response.data:
                print('---Response---')
                print(f'Status Code: {response.status_code}')
                print_json(response.data)

        response_type = ''
        if 'type' in operation_schemas[s['operation_id']]['responses'][str(s['status_code'])].keys():
            response_type = operation_schemas[s['operation_id']]['responses'][str(s['status_code'])]['type']

        if response_type == 'object' or response_type == 'array':
            response.data = json.loads(response.data)

        verify_response(response, s['status_code'], response_schema)
        if verbose:
            print("[bold green]--Step Verified--[/bold green]")

        steps[s['name']] = {}
        steps[s['name']]['response'] = response



import pytest

from src.errors import *

# Test Values
operation_set: set[str] = {"op_one", "op_two", "op_three"}
coverage = 90.0
desired_coverage = 100.0
file_extension = ".json"
base_context_value = "not env or steps"
evaluation_error_message = "evaluation error"
environment_variable_value = "NON_EXISTENT_ENV"
expression_value = "2 + 2 = fish"
schema_value = "application/xml"


def test_undocumented_operation_error():
    with pytest.raises(UndocumentedOperationError) as err:
        raise UndocumentedOperationError(operation_set)


def test_insufficient_coverage_error():
    with pytest.raises(InsufficientCoverageError) as err:
        raise InsufficientCoverageError(coverage, desired_coverage, operation_set)


def test_invalid_file_error():
    with pytest.raises(InvalidFileError) as err:
        raise InvalidFileError(file_extension)


def test_base_context_error():
    with pytest.raises(BaseContextError) as err:
        raise BaseContextError(base_context_value)


def test_evaluation_error():
    with pytest.raises(EvaluationError) as err:
        raise EvaluationError(evaluation_error_message)


def test_environment_variable_error():
    with pytest.raises(EnvironmentVariableError) as err:
        raise EnvironmentVariableError(environment_variable_value)


def test_invalid_expression_error():
    with pytest.raises(InvalidExpressionError) as err:
        raise InvalidExpressionError(expression_value)


def test_unsupported_schema_error():
    with pytest.raises(UnsupportedSchemaError) as err:
        raise UnsupportedSchemaError(schema_value)



from typing import Union

class Response:
    def __init__(
        self,
        status_code: int,
        headers: dict,
        data: Union[dict, str],
    ):
    
        self.status_code = status_code
        self.headers = headers
        self.data = data



import pytest
from hamcrest import assert_that, is_, has_items, instance_of

from src.parsing import *
from src.preprocessing import *

class TestPreProcessing:
    """Class for basic unit testing of the preprocessing.py module
    """
    steps_file_path="./tests/fixtures/valid/steps/person_api.yml"
    spec_file_path = "./tests/fixtures/valid/specs/person_api.yml"

    steps = parse_steps_file(steps_file_path)
    spec, operation_schemas = parse_spec_file(steps, spec_file_path)

    def test_get_operation_coverage(self):
        step_coverage, spec_coverage = get_operation_coverage(self.steps, self.spec)
        assert_that(step_coverage == spec_coverage)

    def test_check_operation_coverage(self):
        check_operation_coverage(self.steps, self.spec)

    # def test_evaluate(expression: any, steps={}):
    #     return



import os
import re

from .errors import (
    InsufficientCoverageError,
    UndocumentedOperationError,
    BaseContextError,
    EvaluationError,
    EnvironmentVariableError
)
from rich import print

def get_operation_coverage(steps: dict, spec: dict):
    steps_operations = set()
    for operation in steps['steps']:
        steps_operations.add(operation['operation_id'])

    spec_operations = set()
    for path in spec['paths']:
        for method in spec['paths'][path]:
            op_id = spec['paths'][path][method]['operationId']
            spec_operations.add(op_id)

    return steps_operations, spec_operations


def check_operation_coverage(steps: dict, spec: dict):
    steps_operations, spec_operations = get_operation_coverage(steps, spec)

    covered = spec_operations & steps_operations
    uncovered = spec_operations - steps_operations
    undocumented = steps_operations - spec_operations

    proportion_covered = len(covered) / (len(covered) + len(uncovered) + len(undocumented))

    if len(undocumented) > 0:
        has_undocumented_operations = True
    else:
        has_undocumented_operations = False

    # If any undocumented operations, immediately halt
    print("--Checking for Uncovered Operations--")
    if has_undocumented_operations:
        raise UndocumentedOperationError(undocumented)

    # Check if operation coverage meets threshold
    print("--Validating Coverage Threshold--")
    if "coverage_threshold" in steps:
        target_coverage: float = steps["coverage_threshold"]

        if proportion_covered < target_coverage:
            raise InsufficientCoverageError(
                proportion_covered, target_coverage, uncovered
            )

    print("[bold green]--Coverage Threshold Met--[/bold green]")


def evaluate(expression: any, steps={}) -> any:
    """
    Recursively evaluate nested expressions using depth-first search.
    Eventually the evaluation result as a string is returned.
    The only allowed base contexts are "env" and "steps".
    Args:
        expression (str): Object of any type that may contain expression(s)
    Raises:
        EnvironmentVariableError:
            if the expression represents an environment variable but it cannot be found
    Returns:
        The evaluated result as a string if there is any expression, original value otherwise.
    """

    if expression is None:
        return

    # Evaluate each value in a dictionary
    if isinstance(expression, dict):
        return dict(map(lambda x: (x[0], evaluate(x[1], steps)), expression.items()))

    # Evaluate each element in a list
    if isinstance(expression, list):
        return list(map(lambda x: evaluate(x, steps), expression))

    if not isinstance(expression, str):
        return expression

    matches: list[str] = re.findall(r"(\${{[^/}]*}})", expression)
    if not matches:
        return expression

    for match in matches:
        value = match.removeprefix("${{").removesuffix("}}").strip()
        base = value.split(".").pop(0)

        if base == "env":
            # Only split at the first dot
            result = os.environ.get(value.split(".", 1)[1])
            if result is None:
                raise EnvironmentVariableError(value)
        elif base == "steps":
            value_array = value.split(".")
            value_dict = f"{value_array[0]}['{value_array[1]}']['{value_array[2]}'].{value_array[3]}"
            for idx, val in enumerate(value_array):
                if idx > 3:
                    value_dict += f"['{val}']"
            try:
                result = eval(value_dict)
            except AttributeError as e:
                raise EvaluationError(e)
        else:
            raise BaseContextError(base)

        # Only replace the first occurrence
        expression = expression.replace(match, str(result), 1)

    return expression



from rich import print

# TODO: Extend jsonschema.ValidationError class

class UndocumentedOperationError(BaseException):
    """
    Raised when the step file contains operations that are not documented in the OpenAPI spec.
    This error will halt further execution of the action.
    """

    def __init__(self, undocumented: set[str]):
        super().__init__(
            print(f"The following operations from the steps file are undocumented: [bold red]{undocumented}[/bold red]")
        )


class InsufficientCoverageError(BaseException):
    """
    Raised if the step file does not sufficiently cover the OpenAPI spec according the the coverage threshold.
    If a coverage threshold given in the step file, this error will halt further execution of the action.
    """

    def __init__(
        self, achieved_coverage: float, target_coverage: float, uncovered: set[str]
    ):
        super().__init__(
            print(f'''The operation coverage is [bold red]{achieved_coverage}[/bold red]
but the target is [green]{target_coverage}[/green].\n
The following operations are uncovered: [bold red]{uncovered}[/bold red]''')
        )


class InvalidFileError(BaseException):
    """
    Raised when you pass in an API spec file that is not JSON or YAML
    """
    def __init__(self, extension):
        super().__init__(
            print(f"Incorrect type for the API Spec file. Only [green]JSON[/green] and [green]YAML[/green] are supported. [bold red]{extension}[/bold red] supplied.")
        )


class BaseContextError(Exception):
    """
    Raises when the base context of an expression is not either "env" or "steps".
    """
    def __init__(self, value):
        super().__init__(
            f"The base context must be either 'env' or 'steps', but found [bold red]'{value}'[/bold red]"
        )


class EvaluationError(Exception):
    """
    Raises when the errors occur during expression evaluation.
    """
    def __init__(self, message):
        super().__init__(message)


class EnvironmentVariableError(Exception):
    """
    Raises when an environment variable is not found.
    """
    def __init__(self, value):
        super().__init__(f"environment variable {value} not found")


class InvalidExpressionError(Exception):
    """
    Raises when an expression is invalid.
    """
    def __init__(self, value):
        super().__init__(f"invalid expression: {value}")


class UnsupportedSchemaError(Exception):
    """
    Raises when a request body schema is unsupported.
    """
    def __init__(self, value):
        super().__init__(f"{value}\n"
                         f"We do not currently support multiple request/response bodies or content types other than:\n"
                         f"- application/json\n"
                         f"- application/octet-stream")



from jsonschema import validate, ValidationError
from rich import print


def verify_request_body(request, schema):
    try:
        validate(instance=request, schema=schema)
    except ValidationError as e:
        print("[bold red]--Request Validation Failed--[/bold red]")
        raise ValidationError("There was an issue with your request body.") from e
    return


def verify_response(response, status_code, schema):
    if response.status_code != status_code:
        print("[bold red]--Response Validation Failed--[/bold red]")
        raise ValidationError("HTTP Status Code does not match the expected value from the step file.")
    else:
        try:
            validate(instance=response.data, schema=schema)
        except ValidationError as e:
            print("[bold red]--Response Validation Failed--[/bold red]")
            print(e)
            raise ValidationError("There was an issue with the response from the API.")
    return



from setuptools import setup, find_packages
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

with open("requirements.txt", "r") as fp:
    requirements = fp.readlines()

setup(
    name='pypony',
    description='A python utility for contract testing APIs',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Bandwidth',
    author_email='letstalk@bandwidth.com',
    url='https://github.com/Bandwidth/pypony/',
    version='1.0.0b2',
    py_modules=['pypony', 'src'],
    install_requires=requirements,
    packages=find_packages(exclude=["test"]),
    include_package_data=True,
    entry_points='''
        [console_scripts]
        pypony=pypony:cli
    '''
)
